;==============================================================================
    include \masm32\include\masm32rt.inc
;==============================================================================

printf MACRO format:REQ, args:VARARG
    IFNB <args>
        invoke crt_printf, cfm$(format), args
    ELSE
        invoke crt_printf, cfm$(format)
    ENDIF
    EXITM <>
ENDM

;==============================================================================
    .data

      dwPageSize          dd 0
      pvBase              dd 0
      pNextPage           dd 0
      prevExceptionFilter dd 0

      mbi                 MEMORY_BASIC_INFORMATION <>

    .code
;==============================================================================

LPEXCEPTION_POINTERS typedef ptr EXCEPTION_POINTERS
EXCEPTION_GUARD_PAGE equ 080000001h

GuardPageExceptionFilter proc lpexception_pointers:LPEXCEPTION_POINTERS

    mov edx, lpexception_pointers
    mov ecx, [edx].EXCEPTION_POINTERS.pExceptionRecord

    ;-----------------------------------------
    ; Handle the guard page exception and let
    ; the syatem handle any other exception.
    ;-----------------------------------------

    .if [ecx].EXCEPTION_RECORD.ExceptionCode == EXCEPTION_GUARD_PAGE

      ;--------------------------------------------------------------
      ; At this point the system has cleared the PAGE_GUARD modifier
      ; so the page that was formerly the guard page is now a normal
      ; PAGE_READWRITE page, and we need to commit a new guard page.
      ;--------------------------------------------------------------

      invoke VirtualAlloc, pNextPage, dwPageSize , MEM_COMMIT,
                                      PAGE_READWRITE or PAGE_GUARD
      mov eax, dwPageSize
      add pNextPage, eax

      ;--------------------------------------------
      ; This return value will cause execution to
      ; continue from the point of the exception.
      ;--------------------------------------------

      return EXCEPTION_CONTINUE_EXECUTION

    .else

      return EXCEPTION_EXECUTE_HANDLER

    .endif

GuardPageExceptionFilter endp

;==============================================================================

InitSelfExpandingBuffer proc pageLimit:DWORD

    LOCAL sysinfo:SYSTEM_INFO

    ;--------------------------------
    ; Allow room for the guard page.
    ;--------------------------------

    inc pageLimit

    ;----------------------------------------------------
    ; Get and store the page size (normally 4096 bytes).
    ;----------------------------------------------------

    invoke GetSystemInfo, addr sysinfo
    mov eax, sysinfo.dwPageSize
    mov dwPageSize, eax

    ;-------------------------------
    ; Calc the page limit in bytes.
    ;-------------------------------

    mul pageLimit
    push eax
    printf( "reserved %d bytes\n", eax )
    pop eax

    ;------------------------------------------------
    ; Reserve the memory and store the base address.
    ; If any allocation fails here, return 0.
    ;------------------------------------------------

    invoke VirtualAlloc, 0, eax, MEM_RESERVE, PAGE_NOACCESS
    mov pvBase, eax
    .if eax == 0
        return 0
    .endif
    printf( "pvBase %X\n", eax )

    push pvBase
    pop  pNextPage

    ;---------------------------------
    ; Commit one PAGE_READWRITE page.
    ;---------------------------------

    invoke VirtualAlloc, pNextPage, dwPageSize , MEM_COMMIT, PAGE_READWRITE

    .if eax == 0
        return 0
    .endif

    mov eax, dwPageSize
    add pNextPage, eax

    ;------------------------
    ; Commit one guard page.
    ;------------------------

    invoke VirtualAlloc, pNextPage, dwPageSize , MEM_COMMIT,
                                   PAGE_READWRITE or PAGE_GUARD

    .if eax == 0
        return 0
    .endif

    mov eax, dwPageSize
    add pNextPage, eax

    ;---------------------------------------------------------------------
    ; Replace the existing top-level exception filter with our guard page
    ; exception filter, so when an unhandled exception reaches the system
    ; UnhandledExceptionFilter function, and the process is not being
    ; debugged, the function will call our filter procedure.
    ;---------------------------------------------------------------------

    invoke SetUnhandledExceptionFilter, GuardPageExceptionFilter
    mov prevExceptionFilter, eax

    ;---------------------------------
    ; Return a pointer to the buffer.
    ;---------------------------------

    return pvBase

InitSelfExpandingBuffer endp

;==============================================================================

FreeSelfExpandingBuffer proc

    ;----------------------------------------------------------------------
    ; Restore the previous top-level exception filter and free the buffer.
    ;----------------------------------------------------------------------

    invoke SetUnhandledExceptionFilter, prevExceptionFilter
    invoke VirtualFree, pvBase, 0, MEM_RELEASE

    ret

FreeSelfExpandingBuffer endp

;==============================================================================
start:
;==============================================================================

    ;inkey "Press any key to continue..."

    invoke InitSelfExpandingBuffer, 100   ; <- reserve size in pages
    mov esi, eax
    printf( "%X\n", esi )

    N=0
    REPEAT 200
        mov [esi+N], eax
        invoke VirtualQuery, esi, addr mbi, sizeof MEMORY_BASIC_INFORMATION
        printf( "%X\t%X\n", mbi.BaseAddress, mbi.RegionSize )
        N=N+1000
        ;inkey "Press any key to continue..."
    ENDM

    invoke FreeSelfExpandingBuffer

    inkey "Press any key to exit..."
    exit

;==============================================================================
end start
